package de.uni_passau.fim.seibt.gitwrapper.repo;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.ExecRes;

/**
 * A git {@link Repository}.
 */
public class Repository {

    private static final Logger LOG = Logger.getLogger(Repository.class.getCanonicalName());
    private static final String FAILURE_PREFIX = "fatal";

    private GitWrapper git;

    private String url;
    private File dir;

    private Map<String, Commit> commits;

    /**
     * Constructs a new {@link Repository}.
     *
     * @param git
     *         the {@link GitWrapper} to use
     * @param url
     *         the url the {@link Repository} was cloned from
     * @param dir
     *         the directory containing the {@link Repository}
     */
    Repository(GitWrapper git, String url, File dir) {
        this.git = git;
        this.url = url;

        try {
            this.dir = dir.getCanonicalFile();
        } catch (IOException e) {
            LOG.log(Level.WARNING, e, () -> "Exception computing the canonical file for " + dir + ". Using the absolute file instead.");
            this.dir = dir.getAbsoluteFile();
        }

        this.commits = new HashMap<>();
    }

    /**
     * Returns the merge commits in this repository.
     *
     * @return the merge commits
     */
    public List<Commit> getMergeCommits() {
        Optional<ExecRes> revList = git.exec(dir, "rev-list", "--all", "--merges");
        Function<ExecRes, List<Commit>> toCommitList = res -> {
            String[] lines = res.output.split("[\\r?\\n]+");

            LOG.fine(() -> String.format("Found %d merge commits in %s.", lines.length, this));
            LOG.finer(() -> String.format("Merge commits are:%n%s", String.join(System.lineSeparator(), lines)));

            return Arrays.stream(lines).map(this::getCommit).collect(Collectors.toList());
        };

        return revList.map(toCommitList).orElse(Collections.emptyList());
    }

    /**
     * Returns the merge base of the two given {@link Commit} objects. Both must be from the same (this)
     * {@link Repository}.
     *
     * @param a
     *         the first {@link Commit}
     * @param b
     *         the second {@link Commit}
     * @return the merge base or
     */
    public Optional<Commit> mergeBase(Commit a, Commit b) {

        if (!(commits.values().contains(a) && commits.values().contains(b))) {
            LOG.warning(() -> "Both commits must be part of the repository to get a merge base for them.");
            return Optional.empty();
        }

        Optional<ExecRes> mergeBase = git.exec(dir, "merge-base", a.getId(), b.getId());
        Function<ExecRes, Commit> toCommit = res -> {

            if (res.output.startsWith(FAILURE_PREFIX)) {
                return null;
            }

            Commit base = getCommit(res.output.trim());

            LOG.fine(() -> String.format("Commits %s and %s have the merge base %s.", a.getId(), b.getId(), base.getId()));

            return base;
        };

        return mergeBase.map(toCommit);
    }

    /**
     * Returns a {@link Commit} for the given id. The <code>id</code> is not checked for validity, this method
     * only ensures that only one {@link Commit} object is created for every <code>id</code>.
     *
     * @param id the ID of the commit
     * @return the {@link Commit}
     */
    public Commit getCommit(String id) {
        return commits.computeIfAbsent(id, theID -> new Commit(this, theID));
    }

    /**
     * Returns the {@link GitWrapper} used by this {@link Repository}.
     *
     * @return the {@link GitWrapper}
     */
    public GitWrapper getGit() {
        return git;
    }

    /**
     * Returns the URL this {@link Repository} was cloned from.
     *
     * @return the URL
     */
    public String getUrl() {
        return url;
    }

    /**
     * Returns the directory containing this {@link Repository}.
     *
     * @return the repository directory
     */
    public File getDir() {
        return dir;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        Repository that = (Repository) o;
        return Objects.equals(url, that.url) && Objects.equals(dir, that.dir);
    }

    @Override
    public int hashCode() {
        return Objects.hash(url, dir);
    }

    @Override
    public String toString() {
        return String.format("Repository{url=%s, dir=%s}", url, dir);
    }
}
